// ReaderAccess.c
// Routines to access GiS Reader with LowLevel Protocol
// Copyright: GiS Gesellschaft fr Informatik und Steuerungstechnik mbH
// Author:    Martin Blank
//
// The communication to the Transponder Read/Write device is don through
// a serial interface with the following parameters:
// Baudrate:  19200
// Databits:      8
// Stopbits:      1
// Parity  :   none

#ifdef _WIN_DEBUG
#include "windows.h"
char HexChar(unsigned char nibble)
{
  if (nibble < 10)
    return nibble | 0x30;
  else
    return (nibble - 9) | 0x40;
}

void DebugOut(unsigned char ch)
{
  char String[20];
  String[0] = ' ';
  String[1] = HexChar((ch & 0xF0) >> 4);
  String[2] = HexChar((ch & 0x0F));
  String[3] = 0;
  OutputDebugString(String);
}

void DebugNL()
{
  char String[20];
  String[0] = '\n';
  String[1] = 0;
  OutputDebugString(String);
}

void DebugOutBlock(unsigned char * pData, int Len)
{
  while (Len--)
    DebugOut(*pData++);
}

void DebugOutString(char * pString)
{
  OutputDebugString(pString);
}
#else
#define DebugOut(x)
#define DebugNL()
#define DebugOutBlock(x,y)
#define DebugOutString(x)
#endif

#include "ReaderAccess.h"

#define TIMEOUT_TIME  200

#define STX           0x02
#define DEVICE_ADDR   0x01

#define CMD_UID_REQUEST                   0xC0
#define CMD_SELECT                        0xC1
#define CMD_READ_BLOCKS                   0xC2
#define CMD_WRITE_BLOCKS                  0xC3
#define CMD_CUSTOM                        0xC4

#define CMD_GET_VERSION                   0xF0
#define SET_RF                            0xF5

#define CUSTOM_HITAGS_AUTHENTICATE        0x01

#define BUFF_SIZE 64

static unsigned char InBuff[BUFF_SIZE];
static char InBuffRd;
static char InBuffWr;

static unsigned char OutBuff[BUFF_SIZE];

static int LastErr = 0;

static void my_memcpy(void * dest, void * src, int len)
{
  unsigned char * pbDest = (unsigned char*)dest;
  unsigned char * pbSrc  = (unsigned char*)src;
  while (len--)
    *pbDest++ = *pbSrc++;
}

// To write data via the serial interface, the following function has to be provided
// pData:         pointer to data to send
// Len            Length in bytes of data to send
// return value:  1: successfully sent, 0: not sent
int SendPort(unsigned char * pData, int Len);

// To enable timeout recognition a routine for time measurement is needed
// returns actual time in Milliseconds
unsigned int TimeMillisecond(void);

// Allow Task change to activate other program parts
// This function has to be provided, even if if does nothing
void TaskChange(void);

// This function has to be called from the read routine of the serial interface
// to load the data
// ch:            Byte read from interface
void AddData(unsigned char ch)
{
  DebugOut(ch);

  InBuff[InBuffWr] = ch;

  if (InBuffWr > BUFF_SIZE)
    InBuffWr = 0;
  else
    InBuffWr++;
}

static unsigned char CalcCRC(unsigned char * pData, int DataLen)
{
  int i;
  unsigned char CRC = 0;

  for(i = 1; i < DataLen; i++)
  {
    CRC ^= pData[i];
  }
  return CRC;
}

static int SendCommand(int cmd,unsigned char * pData, int DataLen)
{
  int cmdLen;
  int i;

  cmdLen = 0;
  OutBuff[cmdLen++] = STX;
  OutBuff[cmdLen++] = DEVICE_ADDR;
  OutBuff[cmdLen++] = cmd;
  if (pData == 0)
  {
    OutBuff[cmdLen++] = 0;
  }
  else
  {
    OutBuff[cmdLen++] = DataLen;
  
    for (i = 0; i < DataLen; i++)
      OutBuff[cmdLen++] = pData[i];
  }

  OutBuff[cmdLen++] = CalcCRC(OutBuff,cmdLen);
  DebugOutString(">>");
  DebugOutBlock(OutBuff, cmdLen);
  DebugNL();
  return SendPort(OutBuff,cmdLen);
}

static void RecvBuffEmpty()
{
  InBuffWr = 0;
  InBuffRd = 0;
  DebugNL();
}

static int WaitForAnswer(unsigned int Timeout)
{
  int Count;
  unsigned int StartTime = TimeMillisecond();

  // At least 4 byte have to be in a block
  while ((InBuffWr - InBuffRd) < 4)
  {
    TaskChange();
    if ((TimeMillisecond() - StartTime) > Timeout )
      return 0;
  }

  if ( InBuff[InBuffRd] != STX)
    return 0;

  // Count is : Data length + 4 Header bytes + 1 byte checksum
  Count = InBuff[3] + 5;

  StartTime = TimeMillisecond();
  while ((InBuffWr - InBuffRd) < Count)
  {
    TaskChange();
    if ((TimeMillisecond() - StartTime) > Timeout )
      return 0;
  }

  return 1;
}

static void WaitMillisecond(unsigned int Dauer)
{
  unsigned int StartTime = TimeMillisecond();

  while ((TimeMillisecond() - StartTime) < Dauer)
    TaskChange();
}

// Returns last error
int TS_GetLastErr(void)
{
  return LastErr;
}

static char Transfer(unsigned char cmd, unsigned char * pSendData, unsigned char SendLen, unsigned char * pRecvData, unsigned char MaxRecvLen)
{
  char RecvLen;

  RecvBuffEmpty();

  if (SendCommand(cmd,pSendData,SendLen) == 0)
    return -1;

  if (WaitForAnswer(TIMEOUT_TIME) == 0)
  {
    // Timeout Error
    LastErr = TIMEOUT_ERR;
    return -1;
  }

  if (CalcCRC(InBuff,InBuffWr-1) != InBuff[InBuffWr-1])
  {
    // Checksum error
    LastErr = CRC_ERR;
    return -1;
  }

  if (InBuff[2] != 0)
  {
    // Fehlerantwort
    LastErr = InBuff[2];
    return -1;
  }

  RecvLen = InBuff[3];

  if (RecvLen > MaxRecvLen)
    RecvLen = MaxRecvLen;

  if (pRecvData)
    my_memcpy(pRecvData, &InBuff[4], RecvLen);

  LastErr = NO_ERR;
  return RecvLen;
}

int TS_GetVersion(char * pVersion)
{
  return Transfer(CMD_GET_VERSION, 0, 0, pVersion, 4);
}

int TS_UIDRequest(int TTyp, unsigned char * pUID, int MaxUIDLen)
{
  char RecvLen;
  unsigned char SendCmd[3];

  SendCmd[0] = TTyp;
  RecvLen = Transfer(CMD_UID_REQUEST, SendCmd, 1, pUID, MaxUIDLen);
  if (RecvLen < 0)
    return 0;

  if (RecvLen < 4)
  {
    // Not enough data
    LastErr = DATA_SIZE_ERR;
    return 0;
  }

  LastErr = NO_ERR;
  return RecvLen;
}

int TS_Select(int TTyp, unsigned char * pUID,unsigned char * pConfig)
{
  char RecvLen;
  unsigned char SendCmd[7];

  SendCmd[0] = TTyp;
  my_memcpy(&SendCmd[1], pUID, 4);
  RecvLen = Transfer(CMD_SELECT, SendCmd, 5, pConfig, 4);
  if (RecvLen < 0)
    return 0;

  if (RecvLen < 4)
  {
    // Not enough data
    LastErr = DATA_SIZE_ERR;
    return 0;
  }

  LastErr = NO_ERR;
  return 4;
}

int TS_SetRF(unsigned char On)
{
  char RecvLen;

  RecvLen = Transfer(SET_RF, &On, 1, 0, 0);
  if (RecvLen < 0)
    return 0;

  LastErr = NO_ERR;
  return 1;
}

//Read Blocks
int TS_Read(int TTyp, int BlockNr, unsigned char * pData, int Len)
{
  unsigned char SendCmd[3];
  int ReadLen;

  SendCmd[0] = TTyp;
  SendCmd[1] = (unsigned char)BlockNr;
  SendCmd[2] = Len/4; // Number of blocks, Len is given in Bytes, has to be multiple of 4.
  ReadLen = Transfer(CMD_READ_BLOCKS, SendCmd, 3, pData, Len);
  if (ReadLen < 0)
    return 0;

  if (ReadLen%4 != 0)
  {
    // not enough data received
    LastErr = DATA_SIZE_ERR;
    return 0;
  }

  LastErr = NO_ERR;
  return ReadLen;
}

// Write Block (multiples of 4 Byte)
int TS_Write(int TTyp, int BlockNr, unsigned char * pData, int DataLen)
{
  char RecvLen;
  unsigned char SendCmd[30];
  char SendCmdLen;
  int i;

  SendCmdLen = 0;
  SendCmd[SendCmdLen++] = TTyp;
  SendCmd[SendCmdLen++] = (unsigned char)BlockNr;
  for (i = 0;(i < DataLen) && (SendCmdLen < sizeof(SendCmd)); i++)
    SendCmd[SendCmdLen++] = pData[i];

  // Fill with empty Bytes to have complete blocks
  for (;((i%4) != 0) && (SendCmdLen < sizeof(SendCmd)); i++)
    SendCmd[SendCmdLen++] = 0;

  RecvLen = Transfer(CMD_WRITE_BLOCKS, SendCmd, SendCmdLen, 0, 0);
  if (RecvLen < 0)
    return 0;

  LastErr = NO_ERR;
  return 1;
}
